/* Doors for Linux.
   Copyright (C) 1998 Jason Lango

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*
**  bench.c - Benchmark comparing Doors vs. other forms of IPC.
**  Jason Lango <jal@rampant.org>
*/

#include "test.h"
#include <door.h>
#include <stdarg.h>
#include "bench_rpc.h"

#define READ_END 0
#define WRITE_END 1
#define NTIMES 1000

typedef struct bench {
	const char *name;
	void (*init)(void);
	void (*parent_init)(void *data_ptr, size_t data_size);
	void (*parent)(void *data_ptr, size_t data_size);
	void (*child)(void *data_ptr, size_t data_size);
	void (*after_test)(void);
	void (*cleanup)(void);
	size_t max_data;
} bench_t;

bench_t *running_test = NULL;
int child_pid = -1;
int main_pid = -1;

/**********************************************************************
*  Pipe benchmark
*/

int pipe_request[2];
int pipe_reply[2];

void PipeInit(void)
{
	printf("# Pipe read/write, wait for reply\n");
	VERIFY ( pipe(pipe_request) );
	VERIFY ( pipe(pipe_reply) );
}

void PipeParent(void *data_ptr, size_t data_size)
{
	int nbytes;
	int ii;

	for (ii = 0; ii < NTIMES; ii++) {
		if (write(pipe_request[WRITE_END], data_ptr, data_size)
			!= data_size) {
			perror("write didn't write enough");
			abort();
		}
		do {
			VERIFY ( nbytes = read(pipe_reply[READ_END],
				data_ptr, 1) );
		} while (nbytes != 1);
	}
}

void PipeChild(void *data_ptr, size_t data_size)
{
	int nbytes;
	int total;

	for (;;) {
		total = 0;

		do {
			VERIFY ( nbytes = read(pipe_request[READ_END],
				data_ptr, data_size - total) );
			total += nbytes;
		} while (total != data_size);

		VERIFY ( write(pipe_reply[WRITE_END], data_ptr, 1) );
	}
}

void PipeCleanup(void)
{
	VERIFY ( close(pipe_request[0]) );
	VERIFY ( close(pipe_request[1]) );
	VERIFY ( close(pipe_reply[0]) );
	VERIFY ( close(pipe_reply[1]) );
}

/**********************************************************************
*  Door benchmark
*/

const char *bench_door = "bench_door";
const char *initialized = "bench_init";
int door_door = -1;

void DoorServe(void *cookie, char *data_ptr, size_t data_size,
	door_desc_t *desc_ptr, size_t desc_num)
{
	VERIFY ( door_return(NULL, 0, NULL, 0) );
}

void DoorInit(void)
{
	printf("# Door call\n");
}

void DoorCleanup(void)
{
	if (door_door >= 0) {
		VERIFY ( close(door_door) );
		door_door = -1;
	}

	fdetach(bench_door);
	unlink(bench_door);
	unlink(initialized);
}

void DoorParentInit(void *data_ptr, size_t data_size)
{
	int fd;

	/* Wait for the child to initialize, avoiding pesky race-condition.
	*/
	while ((fd = open(initialized, O_RDONLY)) < 0)
		;
	VERIFY ( close(fd) );
	VERIFY ( unlink(initialized) );

	VERIFY ( door_door = open(bench_door, O_RDONLY) );
}

void DoorParent(void *data_ptr, size_t data_size)
{
	int ii;

	for (ii = 0; ii < NTIMES; ii++) {
		door_arg_t args;

		args.data_ptr = data_ptr;
		args.data_size = data_size;
		args.desc_ptr = NULL;
		args.desc_num = 0;
		args.rbuf = NULL;
		args.rsize = 0;

		VERIFY ( door_call(door_door, &args) );
	}
}

void DoorChild(void *data_ptr, size_t data_size)
{
	int door;

	VERIFY ( door = door_create(DoorServe, NULL, 0) );
	VERIFY ( creat(bench_door, 0600) );
	VERIFY ( fattach(door, bench_door) );

	/* Tell the parent that we've initialized.
	*/
	VERIFY ( creat(initialized, 0600) );

	for (;;)
		pause();
}

/**********************************************************************
*  ONC-RPC 
*/

int rpc_srv_pid = 0;
CLIENT *rpc_clnt = NULL;

int bench_rpc_svr(int argc, char *argv[]);

int *rpc_call_1_svc(rpc_inargs *inargs, struct svc_req *req)
{
	static int one = 1;
	/* Do nothing. */
	return &one;
}

void RpcInit(char *trans)
{
	fflush(NULL);

	VERIFY ( rpc_srv_pid = fork() );
	if (!rpc_srv_pid) {
		bench_rpc_svr(0, NULL);
		exit(1);
	}

	/* Give the server a chance to warm up.
	*/
	sleep(2);

	rpc_clnt = clnt_create("localhost", BENCH_RPC, BENCH_RPC_VERS, trans);
	if (!rpc_clnt) {
		printf("failed to create rpc client!\n");
		abort();
	}
}

void RpcUdpInit(void)
{
	printf("# ONC-RPC, using UDP\n");
	RpcInit("udp");
}

void RpcTcpInit(void)
{
	printf("# ONC-RPC, using TCP\n");
	RpcInit("tcp");
}

void RpcParent(void *data_ptr, size_t data_size)
{
	int snork;
	rpc_inargs inargs;

	inargs.rpc_inargs_len = data_size;
	inargs.rpc_inargs_val = data_ptr;

	for (snork = 0; snork < NTIMES; snork++) {
		if (!rpc_call_1(&inargs, rpc_clnt)) {
			clnt_perror(rpc_clnt, "rpc_call");
			abort();
		}
	}
}

void RpcChild(void *data_ptr, size_t data_size)
{
	for (;;)
		pause();
}

void RpcCleanup(void)
{
	if (rpc_srv_pid) {
		VERIFY ( kill(rpc_srv_pid, SIGTERM) );
		rpc_srv_pid = 0;
	}

	if (rpc_clnt) {
		auth_destroy(rpc_clnt->cl_auth);
		clnt_destroy(rpc_clnt);
		rpc_clnt = NULL;
	}
}

/**********************************************************************/

bench_t tests[] = {
	{ "pipe", PipeInit, NULL, PipeParent, PipeChild, NULL,
		PipeCleanup, 0 },
	{ "door", DoorInit, DoorParentInit, DoorParent, DoorChild,
		DoorCleanup, NULL, 0 },
	{ "rpc-udp", RpcUdpInit, NULL, RpcParent, RpcChild, NULL,
		RpcCleanup, 0 },
	{ "rpc-tcp", RpcTcpInit, NULL, RpcParent, RpcChild, NULL,
		RpcCleanup, 0 },
	{ NULL }
};

double timeit(size_t data_size, bench_t *test)
{
	int child;
	char *data_ptr;

	data_ptr = malloc(data_size);
	if (!data_ptr) {
		printf("ERROR: out of memory\n");
		abort();
	}

	/* memset(data_ptr, 0xCC, data_size); */

	fflush(NULL);
	VERIFY ( child_pid = child = fork() );
	if (child) {
		struct timeval tv_begin, tv_end;

		if (test->parent_init)
			(*test->parent_init)(data_ptr, data_size);

		VERIFY ( gettimeofday(&tv_begin, NULL) );

		(*test->parent)(data_ptr, data_size);

		VERIFY ( gettimeofday(&tv_end, NULL) );
		VERIFY ( kill(-child, SIGTERM) );
		VERIFY ( wait(NULL) );
		free(data_ptr);

		if (test->after_test)
			(*test->after_test)();

		return ((tv_end.tv_sec - tv_begin.tv_sec) * 1.0e6 +
			(tv_end.tv_usec - tv_begin.tv_usec)) / NTIMES;
	}
	else {
		VERIFY ( setpgrp() );
		(*test->child)(data_ptr, data_size);
	}

	/* shouldn't get here */
	abort();
	return -1;
}

void run_one_size(bench_t *test, size_t data_size)
{
	running_test = test;
	(*test->init)();
	printf("# %18s %20s\n", "Data Size", "Time (us)");
	printf("%20ld %20.0f\n",
		(unsigned long) data_size,
		timeit(data_size, test));
	child_pid = -1;
	if (test->cleanup)
		(*test->cleanup)();
	running_test = NULL;
}

void run_one(bench_t *test)
{
	size_t data_size;
	size_t max_data = 1 * 1024 * 1024;

	running_test = test;
	(*test->init)();
	printf("# %18s %20s\n", "Data Size", "Time (us)");

	if (test->max_data)
		max_data = test->max_data;

	data_size = 2;
	for (; data_size < max_data; data_size *= 1.6) {
		printf("%20ld %20.0f\n",
			(unsigned long) data_size,
			timeit(data_size, test));
	}
	if (test->cleanup)
		(*test->cleanup)();
	running_test = NULL;
}

void cleanup(void)
{
	static int cleaned_up = 0;

	if (!cleaned_up) {
		if (getpid() == main_pid) {
			cleaned_up = 1;
			if (running_test) {
				if (running_test->after_test)
					(*running_test->after_test)();
				if (running_test->cleanup)
					(*running_test->cleanup)();
			}
			if (child_pid > 0)
				kill(-child_pid, SIGTERM);
			kill(-getpid(), SIGTERM);
		}
	}
}

void sigcleanup(int signo)
{
	cleanup();
	exit(1);
}

void test_error(const char *fmt, ...)
{
	va_list ap;

	fprintf(stderr, "bench: ");

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);

	fprintf(stderr, "\n");
	fflush(0);
}

int main(int argc, char *argv[])
{
	bench_t *test;

	signal(SIGTERM, sigcleanup);
	signal(SIGHUP, sigcleanup);
	signal(SIGINT, sigcleanup);
	signal(SIGABRT, sigcleanup);
	atexit(cleanup);

	VERIFY ( setpgrp() );
	VERIFY ( main_pid = getpid() );

	if (argc > 1) {
		for (test = tests; test->parent != NULL; test++) {
			if (!strcmp(test->name, argv[1]))
				break;
		}
		if (test->parent == NULL) {
			printf("unknown test %s\n", argv[1]);
			exit(1);
		}

		if (argc > 2)
			run_one_size(test, strtol(argv[2], NULL, 10));
		else
			run_one(test);
		exit(1);
	}

	for (test = tests; test->parent != NULL; test++) {
		run_one(test);
	}
	return 0;
}
