/* Doors for Linux.
   Copyright (C) 1998 Jason Lango

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*
**  vm_copy.c - Copy-on-write an address range between processes.
**  Jason Lango <jal@rampant.org>
*/

#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/smp_lock.h>
#include <linux/string.h>
#include <linux/swap.h>
#include <linux/types.h>
#include <asm/uaccess.h>
#include <asm/pgtable.h>
#include "doors_debug.h"
#include "vm_copy.h"

static const char *vmcp_bad_pte_printk =
	"doors_vmcp: %s: page fault didn't fill in pte!\n";

static inline int put_pte(struct mm_struct *mm, unsigned long addr,
	pte_t pte_in)
{
	pgd_t *pgd;
	pmd_t *pmd;
	pte_t *pte;

	pgd = pgd_offset(mm, addr);
	pmd = pmd_alloc(pgd, addr);
	if (!pmd)
		goto no_memory;
	pte = pte_alloc(pmd, addr);
	if (!pte)
		goto no_memory;
	set_pte(pte, pte_in);
	return 0;

no_memory:
	return -ENOMEM;
}

static inline int put_page(struct mm_struct *mm, unsigned long addr,
	unsigned long page, struct vm_area_struct *vma)
{
	return put_pte(mm, addr,
		pte_mkwrite(pte_mkdirty(mk_pte(page, vma->vm_page_prot))));
}

static inline int get_pte(struct mm_struct *mm, unsigned long addr,
	pte_t **pte_ret)
{
	pgd_t *pgd;
	pmd_t *pmd;
	pte_t *pte;

	pgd = pgd_offset(mm, addr);
	if (pgd_none(*pgd))
		goto error;
	if (pgd_bad(*pgd))
		goto pgd_bad_error;
	pmd = pmd_offset(pgd, addr);
	if (pmd_none(*pmd))
		goto error;
	if (pmd_bad(*pmd))
		goto pmd_bad_error;
	pte = pte_offset(pmd, addr);

	*pte_ret = pte;
	return 0;

pmd_bad_error:
	printk(KERN_ERR "doors_vmcp: get_pte: bad page middle entry %08lx\n",
		pmd_val(*pmd));
	pmd_clear(pmd);
	return -EFAULT;

pgd_bad_error:
	printk(KERN_ERR "doors_vmcp: get_pte: bad page dir entry %08lx\n",
		pgd_val(*pgd));
	pgd_clear(pgd);
	return -EFAULT;

error:
	return -EFAULT;
}

static inline struct vm_area_struct *page_vma_alloc(struct mm_struct *mm,
	unsigned long addr, unsigned long n_pages)
{
	struct vm_area_struct *vma;

	vma = kmem_cache_alloc(vm_area_cachep, SLAB_KERNEL);
	if (!vma)
		goto vma_alloc_error;

	vma->vm_mm = mm;
	vma->vm_start = addr;
	vma->vm_end = addr + PAGE_SIZE * n_pages;
	vma->vm_flags = VM_MAYREAD | VM_READ
		| VM_MAYWRITE | VM_WRITE
		| VM_MAYEXEC | VM_EXEC
		| mm->def_flags;
	vma->vm_page_prot = protection_map[vma->vm_flags & 0x0f];
	vma->vm_ops = NULL;
	vma->vm_offset = 0;
	vma->vm_file = NULL;
	vma->vm_pte = 0;

	insert_vm_struct(mm, vma);
	return vma;

vma_alloc_error:
	return NULL;
}

static inline void fault_in_readable(struct task_struct *tsk,
	struct vm_area_struct *vma, unsigned long addr)
{
	dprintk(DEBUG_VM_COPY_FAULT, ("doors_vmcp: fault_in_readable: "
		"task 0x%lx, addr 0x%lx, vma 0x%lx\n",
		(unsigned long) tsk,
		(unsigned long) addr,
		(unsigned long) vma));
	handle_mm_fault(tsk, vma, addr, 0);
}

static inline int map_end(struct mm_struct *mm_dst,
	struct mm_struct *mm_src,
	unsigned long *dst_addr, unsigned long *src_addr,
	unsigned long *len, unsigned long src_offset,
	struct vm_area_struct **src_vma,
	struct task_struct *tsk_src)
{
	pte_t *src_pte;
	unsigned long src_page;
	unsigned long dst_page;
	struct vm_area_struct *vma;
	int error;
	unsigned long copy_size;

	if (!(*src_vma))
		return -EFAULT;
	if ((*src_vma)->vm_start > *src_addr)
		return -EFAULT;

	dprintk(DEBUG_VM_COPY, ("doors_vmcp: map_end: copying %lx -> %lx, "
		"len %ld, offset %ld\n",
		*src_addr, *dst_addr, *len, src_offset));

	error = -ENOMEM;
	dst_page = get_free_page(GFP_USER);
	if (!dst_page)
		goto out_error;

	error = get_pte(mm_src, *src_addr, &src_pte);
	if (error)
		goto out_free_page;

	/* If the page is not in memory, fault it in.
	*/
	if (!pte_present(*src_pte))
		fault_in_readable(tsk_src, *src_vma, *src_addr);
	/* Sanity check. */
	if (!pte_present(*src_pte))
		goto bad_pte;

	src_page = pte_page(*src_pte);

	error = -ENOMEM;
	vma = page_vma_alloc(mm_dst, *dst_addr, 1);
	if (!vma)
		goto out_free_page;

	copy_size = (*len > PAGE_SIZE ? PAGE_SIZE : *len) - src_offset;

	memcpy((void*) (dst_page + src_offset),
		(void*) (src_page + src_offset), copy_size);

	error = put_page(mm_dst, *dst_addr, dst_page, vma);
	if (error)
		goto out_free_page;

	dprintk(DEBUG_VM_COPY, ("doors_vmcp: map_end: new page at %lx\n",
		(unsigned long) (*dst_addr & PAGE_MASK)));

	*dst_addr += PAGE_SIZE;
	*src_addr += PAGE_SIZE;
	*len -= PAGE_SIZE;

	if ((*src_vma)->vm_end <= *src_addr)
		*src_vma = (*src_vma)->vm_next;

	return 0;

bad_pte:
	printk(vmcp_bad_pte_printk, "map_end");
	return -ENOMEM;

out_free_page:
	free_page(dst_page);
out_error:
	dprintk(DEBUG_VM_COPY, ("doors_vmcp: map_end: exiting with error %d\n",
		(int) error));
	return error;
}

static inline int map_alloc(struct mm_struct *mm, unsigned long addr,
	unsigned long len)
{
	unsigned long n_pages;
	struct vm_area_struct *vma;

	/* Assumes that addr is page aligned.
	*/
	n_pages = PAGE_ALIGN(len) >> PAGE_SHIFT;
	vma = page_vma_alloc(mm, addr, n_pages);
	if (!vma)
		return -ENOMEM;

	dprintk(DEBUG_VM_COPY, ("doors_vmcp: allocated %ld pages\n",
		(unsigned long) n_pages));
	return 0;
}

static inline int vmcp_copy_page(struct mm_struct *mm_dst,
	struct mm_struct *mm_src,
	unsigned long dst_addr, unsigned long src_addr,
	struct vm_area_struct *vma, struct task_struct *tsk_src,
	struct vm_area_struct *src_vma)
{
	pte_t *src_pte;
	int error;
	unsigned long src_page;
	unsigned long dst_page;

	/* XXX: Should check if the page is the special ZERO_PAGE,
	 * then really just copy-on-write.
	 */

	error = -ENOMEM;
	dst_page = __get_free_page(GFP_USER);
	if (!dst_page)
		goto out_error;

	error = get_pte(mm_src, src_addr, &src_pte);
	if (error)
		goto out_free_page;

	if (!pte_present(*src_pte))
		fault_in_readable(tsk_src, src_vma, src_addr);
	/* Sanity check. */
	if (!pte_present(*src_pte))
		goto bad_pte;

	src_page = pte_page(*src_pte);

	copy_page(dst_page, src_page);

	error = put_page(mm_dst, dst_addr, dst_page, vma);
	if (error)
		goto out_free_page;
	return 0;

bad_pte:
	printk(vmcp_bad_pte_printk, "vmcp_copy_page");
	return -ENOMEM;

out_free_page:
	free_page(dst_page);
out_error:
	return error;
}

static inline int vmcp_copy_page_cow(struct mm_struct *mm_dst,
	struct mm_struct *mm_src,
	unsigned long dst_addr, unsigned long src_addr,
	struct task_struct *tsk_src, struct vm_area_struct *src_vma)
{
	pgd_t *dst_pgd, *src_pgd;
	pmd_t *dst_pmd, *src_pmd;
	pte_t *dst_pte, *src_pte;
	unsigned long page_nr;
	pte_t pte;

	/* Lookup pte in source, allocating page table space as
	 * appropriate.
	 */
	src_pgd = pgd_offset(mm_src, src_addr);
	src_pmd = pmd_alloc(src_pgd, src_addr);
	if (!src_pmd)
		return -ENOMEM;
	src_pte = pte_alloc(src_pmd, src_addr);
	if (!src_pte)
		return -ENOMEM;

	/* If there's no mapping, fault in the page.
	*/
	if (pte_none(*src_pte))
		fault_in_readable(tsk_src, src_vma, src_addr);
	/* Sanity check. */
	if (pte_none(*src_pte))
		goto bad_pte;

	/* Lookup pte in destination, allocating page table space as
	 * appropriate.
	 */
	dst_pgd = pgd_offset(mm_dst, dst_addr);
	dst_pmd = pmd_alloc(dst_pgd, dst_addr);
	if (!dst_pmd)
		return -ENOMEM;
	dst_pte = pte_alloc(dst_pmd, dst_addr);
	if (!dst_pte)
		return -ENOMEM;

	/* Copy one pte from source to destination.  See also:
	 * linux/mm/memory.c, copy_page_range(), copy_one_pte.
	 */

	pte = *src_pte;

	/* If the page is swapped out, deal.  This isn't the average
	 * case, so we'll push it to the bottom of the function.
	 */
	if (!pte_present(pte))
		goto not_present;

	page_nr = MAP_NR(pte_page(pte));
	if (page_nr >= max_mapnr || PageReserved(mem_map + page_nr)) {
		set_pte(dst_pte, pte);
		return 0;
	}

	pte = pte_wrprotect(pte);
	set_pte(dst_pte, pte_mkold(pte));
	set_pte(src_pte, pte);
	atomic_inc(&mem_map[page_nr].count);
	return 0;

not_present:
	/* Handle the case where the page is in swap.
	*/
	swap_duplicate(pte_val(pte));
	set_pte(dst_pte, pte);
	return 0;

bad_pte:
	printk(vmcp_bad_pte_printk, "vmcp_copy_page_cow");
	return -ENOMEM;
}

static inline int map_middle(struct mm_struct *mm_dst,
	struct mm_struct *mm_src,
	unsigned long *dst_addr, unsigned long *src_addr,
	unsigned long *len, struct vm_area_struct *src_vma,
	struct task_struct *tsk_src)
{
	unsigned long dst_end;
	int error;
	struct vm_area_struct *vma;
	unsigned long nr_pages;

	dst_end = (*dst_addr + *len) & PAGE_MASK;
	nr_pages = (dst_end - *dst_addr) >> PAGE_SHIFT;

	dprintk(DEBUG_VM_COPY, ("doors_vmcp: mapping %ld middle pages\n",
		(unsigned long) nr_pages));

	error = -ENOMEM;
	vma = kmem_cache_alloc(vm_area_cachep, SLAB_KERNEL);
	if (!vma)
		goto out_error;

	vma->vm_mm = mm_dst;
	vma->vm_start = *dst_addr;
	vma->vm_end = dst_end;
	vma->vm_pte = 0;
	vma->vm_flags = VM_MAYREAD | VM_READ
		| VM_MAYWRITE | VM_WRITE
		| VM_MAYEXEC | VM_EXEC;
	vma->vm_page_prot = protection_map[vma->vm_flags & 0x0f];
	vma->vm_file = NULL;
	vma->vm_offset = 0;
	vma->vm_ops = NULL;

	insert_vm_struct(mm_dst, vma);

	dprintk(DEBUG_VM_COPY, ("doors_vmcp: map_middle: mapping %lx -> %lx in "
		"target task\n",
		vma->vm_start, vma->vm_end));

	while (nr_pages) {
		unsigned long nr_pages_vma;
		int real_copy;

		if (!src_vma)
			return -EFAULT;
		if (src_vma->vm_start > *src_addr)
			return -EFAULT;

		nr_pages_vma = (src_vma->vm_end - *src_addr) >> PAGE_SHIFT;
		if (nr_pages_vma > nr_pages)
			nr_pages_vma = nr_pages;

		real_copy = src_vma->vm_flags & VM_SHARED ||
			src_vma->vm_flags & VM_SHM ||
			src_vma->vm_file;

		dprintk(DEBUG_VM_COPY, ("doors_vmcp: %s copy of %ld pages, "
			"src_addr %lx, dst_addr %lx\n",
			(real_copy ? "real" : "cow"),
			nr_pages_vma,
			*src_addr, *dst_addr));

		nr_pages -= nr_pages_vma;

		/* The following pages all come from the same source vma,
		   so they will be mapped with the same flags and we only
		   need to check the permissions once.
		*/
		while (nr_pages_vma--) {
			if (real_copy)
				error = vmcp_copy_page(mm_dst, mm_src,
					*dst_addr, *src_addr, vma,
					tsk_src, src_vma);
			else
				error = vmcp_copy_page_cow(mm_dst, mm_src,
					*dst_addr, *src_addr,
					tsk_src, src_vma);
			if (error)
				goto out_error;

			*dst_addr += PAGE_SIZE;
			*src_addr += PAGE_SIZE;
		}

		src_vma = src_vma->vm_next;
	}

	*len &= ~PAGE_MASK;
	return 0;

out_error:
	return error;
}

static inline size_t get_page_rounded_size(unsigned long addr,
	unsigned long len)
{
	unsigned long start = addr;
	unsigned long end = addr + len;
	return PAGE_ALIGN(end) - (start & PAGE_MASK);
}

int door_vm_copy(struct task_struct *tsk_dst, struct task_struct *tsk_src,
	unsigned long src_addr, unsigned long src_len, unsigned long alloc_len,
	unsigned long *addr_ret)
{
	unsigned long dst_addr;
	unsigned long rounded_len;
	unsigned long src_offset;
	struct vm_area_struct *src_vma;
	int error;
	unsigned long src_addr_it, dst_addr_it;
	struct mm_struct *mm_dst, *mm_src;

	mm_dst = tsk_dst->mm;
	mm_src = tsk_src->mm;

	dprintk(DEBUG_VM_COPY,
		("door_vm_copy(addr=%lx,len=%ld)\n", src_addr, src_len));

	/* Check for address overflow.
	*/
	if ((src_addr + src_len) < src_addr)
		return -EINVAL;

	rounded_len = get_page_rounded_size(src_addr, src_len + alloc_len);
	dst_addr = get_unmapped_area(0, rounded_len);
	if (!dst_addr)
		return -EOVERFLOW;

	dprintk(DEBUG_VM_COPY, ("doors_vmcp: found unmapped area: %lx -> %lx\n",
		(unsigned long) dst_addr,
		(unsigned long) dst_addr + rounded_len));

	src_offset = src_addr & ~PAGE_MASK;
	src_addr = src_addr & PAGE_MASK;
	src_len += src_offset;

	/* Now, src_addr and dst_addr point to the first page boundary
	   of the area to be mapped.  src_offset is the offset of the
	   src address into the first page.
	   src_len is the number of bytes to the end of the source area,
	   from the starting page boundary.
	*/

	src_vma = find_vma(mm_src, src_addr);
	src_addr_it = src_addr;
	dst_addr_it = dst_addr;

	/* Take the mmap lock, in case allocations or
	 * handle_mm_fault() need to block.
	 */
	down(&mm_src->mmap_sem);
	down(&mm_dst->mmap_sem);

	flush_cache_range(mm_dst, dst_addr, rounded_len);

	if (src_offset) {
		error = map_end(mm_dst, mm_src, &dst_addr_it, &src_addr_it,
			&src_len, src_offset, &src_vma, tsk_src);
		if (error)
			goto bad_beginning;
	}

	if (src_len >= PAGE_SIZE) {
		error = map_middle(mm_dst, mm_src, &dst_addr_it, &src_addr_it,
			&src_len, src_vma, tsk_src);
		if (error)
			goto bad_middle;
	}

	if (src_len) {
		error = map_end(mm_dst, mm_src, &dst_addr_it, &src_addr_it,
			&src_len, 0, &src_vma, tsk_src);
		if (error)
			goto bad_end;
	}

	if (alloc_len) {
		error = map_alloc(mm_dst, dst_addr_it, alloc_len);
		if (error)
			goto bad_alloc;
	}

	flush_tlb_range(mm_dst, dst_addr, rounded_len);

	/* Note that we have to release the mmap_sem before calling
	 * merge_segments, since it tries to take the mmap_sem again.
	 */
	up(&mm_dst->mmap_sem);
	up(&mm_src->mmap_sem);

	merge_segments(mm_dst, dst_addr, rounded_len);

	*addr_ret = dst_addr + src_offset;
	return 0;

bad_alloc:
bad_end:
bad_middle:
bad_beginning:
	up(&mm_dst->mmap_sem);
	up(&mm_src->mmap_sem);
	do_munmap(dst_addr, rounded_len);
	return error;
}
